Package: snapr
Title: Convenient Snapshot Testing Functions for Packages
Version: 0.1.0
Authors@R: 
    person("Douglas Ezra", "Morrison", , "demorrison@ucdavis.edu", role = c("aut", "cre"))
Description: Provides convenient snapshot testing functions for packages,
    including expect_snapshot_data() for data.frames and
    expect_snapshot_object() for any R object.
License: MIT + file LICENSE
Language: en-US
Encoding: UTF-8
URL: https://github.com/d-morrison/snapr,
        https://d-morrison.github.io/snapr/
BugReports: https://github.com/d-morrison/snapr/issues
Depends: R (>= 4.1.0)
Imports: dplyr, jsonlite, readr, testthat, tidyselect, tools, waldo
Suggests: altdoc, bslib, knitr, quarto, rmarkdown, spelling, withr
VignetteBuilder: knitr, quarto
Config/testthat/edition: 3
Config/Needs/website: quarto
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-18 22:19:34 UTC; ezramorrison
Author: Douglas Ezra Morrison [aut, cre]
Maintainer: Douglas Ezra Morrison <demorrison@ucdavis.edu>
Repository: CRAN
Date/Publication: 2026-05-22 20:20:02 UTC
Built: R 4.6.0; ; 2026-05-25 04:02:16 UTC; unix

# Generated by roxygen2: do not edit by hand

export(compare_file_object)
export(darwin_variant)
export(expect_snapshot_data)
export(expect_snapshot_object)
export(platform_variant)
export(save_csv)
export(save_deparse)
export(save_json)
export(save_rds)
export(system_os)
importFrom(dplyr,where)

# snapr.py

[![](https://img.shields.io/pypi/v/snapr?style=flat-square&color=blue)](https://pypi.org/project/snapr/)
[![](https://img.shields.io/github/license/c1m50c/snapr?style=flat-square)](https://github.com/c1m50c/snapr/blob/main/LICENSE)
[![](https://img.shields.io/github/actions/workflow/status/c1m50c/snapr/publish.yml?style=flat-square)](https://github.com/c1m50c/snapr/actions/workflows/publish.yml)

Snapr ([/ˈsnæp ər/](http://ipa-reader.xyz/?text=%CB%88sn%C3%A6p:%C9%99r)) is a library that enables a flexible and frictionless way to render snapshots of maps with overlayed geometries. The backing library is written in Rust, these are bindings to said library.

## Example

```python
from snapr import Geometry, Point, Snapr
import requests

def tile_fetcher(coords: list[tuple[int, int]], zoom: int) -> list[tuple[int, int, bytearray]]:
    tiles = []

    for x, y in coords:
        response = requests.get(
            f"https://a.tile.osm.org/{zoom}/{x}/{y}.png",
            headers={"User-Agent": "snapr.py"},
        )

        tiles.append((x, y, bytearray(response.content)))

    return tiles

snapr = Snapr(tile_fetcher=tile_fetcher, zoom=15)

geometries = [
    Geometry.Point(Point(latitude=41.703811459356196, longitude=-103.34835922605679)),
]

snapshot = snapr.generate_snapshot_from_geometries(geometries=geometries)

with open("example.png", "wb") as image:
    image.write(snapshot)
```

![Example Output](https://github.com/user-attachments/assets/7abd9d2e-ea71-4cfb-81d6-0a03150ddd1e)

## License

Licensed under the [MIT License](https://github.com/c1m50c/snapr/blob/main/LICENSE) found at the root of the repository.

# snapr <img src="man/figures/logo.svg" align="right" height="139" alt="snapr hex sticker" />

<!-- badges: start -->
[![R-CMD-check](https://github.com/d-morrison/snapr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/d-morrison/snapr/actions/workflows/R-CMD-check.yaml)
[![Codecov test coverage](https://codecov.io/gh/d-morrison/snapr/graph/badge.svg)](https://app.codecov.io/gh/d-morrison/snapr)
[![](https://www.r-pkg.org/badges/version/snapr?color=orange)](https://cran.r-project.org/package=snapr)
[![](https://cranlogs.r-pkg.org/badges/grand-total/snapr?color=blue)](https://cran.r-project.org/package=snapr)
[![CRAN checks](https://badges.cranchecks.info/summary/snapr.svg)](https://cran.r-project.org/web/checks/check_results_snapr.html)
[![License: MIT](https://img.shields.io/badge/license-MIT-blue.svg)](https://cran.r-project.org/web/licenses/MIT)
[![](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
<!-- badges: end -->

The goal of snapr is to provide convenient snapshot testing functions for R packages:

- `expect_snapshot_data()` - for data.frames (adapted [with permission](https://github.com/bcgov/ssdtools/issues/379) from the [`{ssdtools}` package](https://CRAN.R-project.org/package=ssdtools))
- `expect_snapshot_object()` - generalizes `expect_snapshot_data()` for use with any R object

## Installation

You can install the released version of snapr from [CRAN](https://cran.r-project.org/) with:

``` r
install.packages("snapr")
```

You can install the development version of snapr from [GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("d-morrison/snapr")
```

### Optional: enhanced RDS diffs with diffviewer

`snapr` works best with the [d-morrison/diffviewer](https://github.com/d-morrison/diffviewer)
fork of [diffviewer](https://github.com/r-lib/diffviewer), which wraps
[diffobj](https://github.com/brodieG/diffobj) to provide rich, visual diffs
when reviewing RDS snapshots via `testthat::snapshot_review()`. Install it
with:

``` r
pak::pak("d-morrison/diffviewer")
```

## Examples

### Snapshot data.frames

``` r
library(snapr)

# In a testthat test file:
test_that("iris data is correct", {
  expect_snapshot_data(iris[1:5, ], name = "iris_sample")
})
```

### Snapshot any R object

``` r
library(snapr)

# Snapshot a list (RDS format by default - works for any R object)
test_that("config object is correct", {
  config <- list(
    name = "my_app",
    version = "1.0.0",
    settings = list(debug = TRUE, timeout = 30)
  )
  expect_snapshot_object(config, name = "config")
})

# Snapshot with JSON format for human-readable diffs
test_that("simple data is correct", {
  data <- list(x = 1:5, y = letters[1:5])
  expect_snapshot_object(data, name = "simple_data", writer = save_json)
})

# Snapshot a model (use RDS for complex objects)
test_that("model structure is correct", {
  model <- lm(mpg ~ wt, data = mtcars)
  expect_snapshot_object(model, name = "model")
})
```

# Snapr

[![](https://img.shields.io/crates/v/snapr?style=flat-square)](https://crates.io/crates/snapr)
[![](https://img.shields.io/github/license/c1m50c/snapr?style=flat-square)](https://github.com/c1m50c/snapr/blob/main/LICENSE)
[![](https://img.shields.io/github/actions/workflow/status/c1m50c/snapr/publish.yml?style=flat-square)](https://github.com/c1m50c/snapr/actions/workflows/publish.yml)

Snapr ([/ˈsnæp ər/](http://ipa-reader.xyz/?text=%CB%88sn%C3%A6p:%C9%99r)) is a library that enables a flexible and frictionless way to render snapshots of maps with overlayed geometries.

## Examples

- [Open Street Maps](https://github.com/c1m50c/snapr/blob/main/examples/open-street-maps/) - Collection of binaries using an OSM tile fetcher.
  - [Point](https://github.com/c1m50c/snapr/blob/main/examples/open-street-maps/src/point/) - Example showing how to draw a point geometry.
  - [Line](https://github.com/c1m50c/snapr/blob/main/examples/open-street-maps/src/line/) - Example showing how to draw a line geometry.
  - [Line String](https://github.com/c1m50c/snapr/blob/main/examples/open-street-maps/src/line_string/) - Example showing how to draw a line string geometry.
  - [Polygon](https://github.com/c1m50c/snapr/blob/main/examples/open-street-maps/src/polygon/) - Example showing how to draw a polygon geometry.
- [Labels](https://github.com/c1m50c/snapr/blob/main/examples/label/) - Example showing how to label a point geometry.
- [SVG](https://github.com/c1m50c/snapr/blob/main/examples/svg/) - Example showing how to draw an SVG on top of a point geometry.
- [Batch](https://github.com/c1m50c/snapr/blob/main/examples/batch/) - Example showing how to use a [`TileFetcher::Batch`](https://docs.rs/snapr/latest/snapr/enum.TileFetcher.html#variant.Batch), as opposed to the usual [`TileFetcher::Individual`](https://docs.rs/snapr/latest/snapr/enum.TileFetcher.html#variant.Individual) variant.

## License

Licensed under the [MIT License](https://github.com/c1m50c/snapr/blob/main/LICENSE) found at the root of the repository.

# snapr.py

Python bindings to the [`snapr`](https://crates.io/crates/snapr) library.

Flexible and frictionless way to render snapshots of maps with stylized geometries.

## Examples

### Open Street Maps

```py
from snapr import Geometry, Point, Snapr
import requests

def tile_fetcher(
    coords: list[tuple[int, int]], zoom: int
) -> list[tuple[int, int, bytearray]]:
    tiles = list()

    for x, y in coords:
        response = requests.get(
            f"https://a.tile.osm.org/{zoom}/{x}/{y}.png",
            headers={"User-Agent": "snapr.py"},
        )

        tiles.append((x, y, bytearray(response.content)))

    return tiles


snapr = Snapr(tile_fetcher=tile_fetcher, zoom=15)

geometries = [
    # Chimney Rock, Nebraska
    # https://www.openstreetmap.org/search?lat=41.703811459356196&lon=-103.34835922605679
    Geometry.Point(Point(latitude=41.703811459356196, longitude=-103.34835922605679)),
    # Chimney Rock Cemetery, Nebraska
    # https://www.openstreetmap.org/search?lat=41.702909695820175&lon=-103.33250120288363
    Geometry.Point(Point(latitude=41.69996628239992, longitude=-103.34170814251178)),
    # Chimney Rock Museum, Nebraska
    # https://www.openstreetmap.org/search?lat=41.702909695820175&lon=-103.33250120288363
    Geometry.Point(Point(latitude=41.702909695820175, longitude=-103.33250120288363)),
]

snapshot = snapr.generate_snapshot_from_geometries(geometries=geometries)

with open("example.png", "wb") as image:
    image.write(snapshot)
```

#' Compare RDS files using waldo for better snapshot review
#' @description
#' This comparison function loads RDS files and compares the deserialized
#' R objects rather than raw bytes. This enables better visualization in
#' snapshot_review() by comparing the actual object structure rather than
#' binary serialization.
#'
#' For meaningful diffs in testthat's snapshot_review(), this function
#' compares the R objects after deserialization using [waldo::compare()].
#'
#' @param old Path to the old (reference) RDS file
#' @param new Path to the new RDS file to compare
#' @param print [logical] whether to print [waldo::compare] output to R console;
#' can become very long for complex objects like [lm]s
#' @inheritDotParams waldo::compare
#' @returns [logical] TRUE if objects are identical, FALSE otherwise
#' @export
#' @examples
#' old_obj <- list(a = 1, b = 2)
#' new_obj <- list(a = 1, b = 3)
#' old_path <- tempfile(fileext = ".rds")
#' new_path <- tempfile(fileext = ".rds")
#' saveRDS(old_obj, old_path)
#' saveRDS(new_obj, new_path)
#' compare_file_object(old_path, new_path)
compare_file_object <- function(old,
                                new,
                                print = FALSE,
                                ...) {
  # Load the RDS files
  old_obj <- readRDS(old)
  new_obj <- readRDS(new)
  # Compare using identical - compares the deserialized objects
  # rather than raw bytes, which is more meaningful for R objects
  # This allows testthat's snapshot_review to properly detect changes
  # in the actual data structure rather than serialization differences
  comparison <- waldo::compare(
    old_obj,
    new_obj,
    x_arg = "actual",
    y_arg = "expected",
    ...
  )
  identical <- length(comparison) == 0 # c.f., `testthat:::expect_waldo_equal_`
  if (!identical && print) {
    c("\nDifferences:", comparison) |>
      paste(collapse = "\n") |>
      cat()
  }
  return(identical)
}

#' Snapshot testing for R objects
#' @description
#' A flexible wrapper around [testthat::expect_snapshot_file()] that allows
#' snapshotting any R object, not just data.frames. This function provides
#' a convenient interface for snapshot testing with sensible defaults for
#' serialization.
#'
#' When using RDS format (the default), snapshots are compared using
#' [waldo::compare()] which provides rich, visual diffs in
#' [testthat::snapshot_review()]. This makes it much easier to review
#' changes to complex R objects.
#'
#' @param x An R object to snapshot. Can be any R object including lists,
#'   models, data.frames, vectors, etc.
#' @param name [character] snapshot name (file extension added
#'   automatically)
#' @param writer [function] Function to write the object to a file.
#'   Default is [save_rds()].
#'   Other options include [save_json()], [save_deparse()], [save_csv()].
#'   Custom writer functions should accept `x` and return a file path.
#' @inheritParams compare_file_object
#' @param tolerance If non-`NULL`, used as threshold for ignoring small
#'   floating point differences when comparing numeric vectors. Only applies
#'   when `writer` produces an RDS file (the default); silently ignored for
#'   text-based formats (JSON, CSV, deparse, etc.). See [waldo::compare()] for
#'   full details.
#' @inheritDotParams testthat::expect_snapshot_file -path -name -compare
#' @returns [NULL] (from [testthat::expect_snapshot_file()])
#' @export
#' @details
#' When using RDS format (the default), snapshots can vary across R versions
#' and platforms even with fixed serialization versions. Consider using
#' `variant = platform_variant()` for RDS snapshots to handle these
#' differences, or use text-based formats like JSON or deparse for more
#' stable snapshots across platforms and versions.
#'
#' The RDS comparison uses [waldo::compare()] internally, which provides
#' rich visual diffs in [testthat::snapshot_review()]. This is particularly
#' useful for complex objects like models, nested lists, or data structures
#' where byte-level comparison would be difficult to interpret.
#' @examples
#' \donttest{
#' # expect_snapshot_object() must be called inside a test_that() block with
#' # testthat 3rd edition active. Outside a test suite, the snapshot is
#' # skipped because there is no reference file to compare against.
#' withr::with_tempdir({
#'   testthat::test_that("snapshot examples", {
#'     testthat::local_edition(3)
#'
#'     # Snapshot a list (RDS format with platform/version variant)
#'     expect_snapshot_object(
#'       list(a = 1, b = 2), name = "config", variant = platform_variant()
#'     )
#'
#'     # Snapshot a model
#'     model <- lm(mpg ~ wt, data = mtcars)
#'     expect_snapshot_object(
#'       model, name = "model", variant = platform_variant()
#'     )
#'
#'     # Snapshot with JSON format (for human-readable diffs)
#'     # Text formats don't need variants
#'     expect_snapshot_object(iris[1:5, ], name = "iris", writer = save_json)
#'
#'     # Snapshot with deparse format
#'     expect_snapshot_object(
#'       list(x = 1:5), name = "simple_list", writer = save_deparse
#'     )
#'   })
#' })
#' }
expect_snapshot_object <- function(x,
                                   name,
                                   writer = save_rds,
                                   print = FALSE,
                                   tolerance = NULL,
                                   ...) {
  path <- writer(x)
  # Get file extension from the path returned by writer
  ext <- tools::file_ext(path)
  # Determine comparison method based on file extension
  # Text-based formats use line-by-line comparison
  # (ignores line-ending differences)
  # RDS files use object comparison with diffobj for better visualization
  # Binary formats use byte-by-byte comparison
  text_extensions <- c("txt", "json", "R", "csv", "md", "yml", "yaml", "xml")

  compare <- if (ext %in% text_extensions) {
    testthat::compare_file_text
  } else if (ext == "rds") {
    function(...) {
      compare_file_object(print = print, tolerance = tolerance, ...)
    }
  } else {
    testthat::compare_file_binary
  }



  testthat::expect_snapshot_file(
    path = path,
    name = paste0(name, ".", ext),
    compare = compare,
    ...
  )
}

#' Save an R object to an RDS file
#' @param x An R object to save
#' @returns [character] Path to the temporary RDS file
#' @keywords internal
#' @export
save_rds <- function(x) {
  path <- tempfile(fileext = ".rds")
  saveRDS(x, path, version = 2)
  path
}

#' Save an R object to a JSON file
#' @param x An R object to save
#' @returns [character] Path to the temporary JSON file
#' @keywords internal
#' @export
save_json <- function(x) {
  path <- tempfile(fileext = ".json")
  jsonlite::write_json(x, path, pretty = TRUE, auto_unbox = TRUE)
  path
}

#' Save an R object to a text file using deparse
#' @param x An R object to save
#' @returns [character] Path to the temporary text file
#' @keywords internal
#' @export
save_deparse <- function(x) {
  path <- tempfile(fileext = ".txt")
  writeLines(deparse(x), path)
  path
}

test_that("expect_snapshot_object works with RDS format (default)", {
  # Create a test object
  test_obj <- list(a = 1:3, b = letters[1:3], c = list(x = 1, y = 2))

  # This should create a snapshot
  # RDS files vary by both OS and R version, so use platform_variant
  expect_snapshot_object(
    test_obj, name = "test_list_rds", variant = platform_variant()
  )

})


test_that("expect_snapshot_object works with RDS format for lm objects", {
  # Create a test object

  lm1 <- lm(Sepal.Length ~ Sepal.Width, data = iris)
  # lm1 <- lm(Sepal.Length ~ Sepal.Width, data = iris, subset = Species == "setosa") # nolint
  # lm2 <- lm(Sepal.Length ~ Sepal.Width, data = iris, subset = Species == "versicolor") # nolint
  expect_snapshot_object(
    lm1,
    name = "lm1_rds",
    tolerance = 1e-8,
    print = TRUE,
    variant = platform_variant()
  )

})

test_that("expect_snapshot_object works with JSON format", {
  # Simple object that works with JSON
  test_obj <- list(x = 1:5, y = c("a", "b", "c"))

  # This should create a JSON snapshot
  expect_snapshot_object(test_obj, name = "test_list_json", writer = save_json)
})

test_that("expect_snapshot_object works with deparse format", {
  # Simple object for deparse
  test_obj <- list(x = 1:5)

  # This should create a text snapshot with deparse
  expect_snapshot_object(
    test_obj, name = "test_list_deparse", writer = save_deparse
  )
})

test_that("save_rds creates an RDS file", {
  test_obj <- list(a = 1, b = 2)
  path <- save_rds(test_obj)

  expect_true(file.exists(path))
  expect_equal(tools::file_ext(path), "rds")

  # Verify we can read it back
  loaded <- readRDS(path)
  expect_equal(loaded, test_obj)
})

test_that("save_json creates a JSON file", {
  test_obj <- list(a = 1, b = 2)
  path <- save_json(test_obj)

  expect_true(file.exists(path))
  expect_equal(tools::file_ext(path), "json")

  # Verify it's valid JSON
  loaded <- jsonlite::fromJSON(path)
  expect_equal(loaded$a, test_obj$a)
  expect_equal(loaded$b, test_obj$b)
})

test_that("save_deparse creates a text file", {
  test_obj <- list(x = 1:3)
  path <- save_deparse(test_obj)

  expect_true(file.exists(path))
  expect_equal(tools::file_ext(path), "txt")

  # Verify the content
  content <- readLines(path)
  expect_true(length(content) > 0)
  expect_true(any(grepl("list", content)))
})

test_that("save_csv creates a CSV file", {
  test_df <- data.frame(x = 1:3, y = letters[1:3])
  path <- save_csv(test_df)

  expect_true(file.exists(path))
  expect_equal(tools::file_ext(path), "csv")

  # Verify we can read it back
  loaded <- readr::read_csv(path, show_col_types = FALSE)
  expect_equal(nrow(loaded), nrow(test_df))
  expect_equal(ncol(loaded), ncol(test_df))
})

test_that("expect_snapshot_data works with data.frames", {
  test_df <- data.frame(
    x = c(1.234567, 2.345678, 3.456789),
    y = letters[1:3]
  )

  # This should create a CSV snapshot with rounded numbers
  expect_snapshot_data(test_df, name = "test_dataframe")
})

test_that("expect_snapshot_data rounds numeric columns", {
  test_df <- data.frame(x = c(1.23456789, 2.34567890))
  path <- save_csv(test_df)

  # The save_csv function is used by expect_snapshot_data
  # But the rounding happens before save_csv is called
  # So we test the rounding logic separately
  fun <- function(x) signif(x, digits = 6)
  rounded_df <- dplyr::mutate(
    test_df, dplyr::across(tidyselect::where(is.numeric), fun)
  )

  expect_equal(rounded_df$x[1], signif(1.23456789, 6))
  expect_equal(rounded_df$x[2], signif(2.34567890, 6))
})

test_that("compare_file_object returns TRUE for identical objects", {
  # Create identical objects
  test_obj <- list(a = 1, b = 2, c = list(x = 1:5))
  old_path <- tempfile(fileext = ".rds")
  new_path <- tempfile(fileext = ".rds")

  saveRDS(test_obj, old_path)
  saveRDS(test_obj, new_path)

  result <- compare_file_object(old_path, new_path)
  expect_true(result)
})

test_that(
  "compare_file_object returns FALSE for different objects",
  {
    # Create different objects
    old_obj <- list(a = 1, b = 2)
    new_obj <- list(a = 1, b = 3)
    old_path <- tempfile(fileext = ".rds")
    new_path <- tempfile(fileext = ".rds")

    saveRDS(old_obj, old_path)
    saveRDS(new_obj, new_path)

    result <- compare_file_object(old_path, new_path, print = FALSE) |>
      expect_false(result)
  }
)

test_that("compare_file_object works with complex objects", {
  # Create complex objects (e.g., models)
  withr::local_seed(123)
  old_model <- lm(mpg ~ wt, data = mtcars)
  new_model <- lm(mpg ~ hp, data = mtcars)

  old_path <- tempfile(fileext = ".rds")
  new_path <- tempfile(fileext = ".rds")

  saveRDS(old_model, old_path)
  saveRDS(new_model, new_path)

  result <- compare_file_object(old_path, new_path, print = FALSE)
  expect_false(result)

  # Test with identical simple objects instead of models
  # (models can have environment differences)
  simple_obj <- list(x = 1:10, y = letters[1:10])
  saveRDS(simple_obj, old_path)
  saveRDS(simple_obj, new_path)
  result2 <- compare_file_object(old_path, new_path)
  expect_true(result2)
})

#' Get the current operating system
#'
#' @description
#' Returns the name of the current operating system in lowercase.
#' This is a simple wrapper around `Sys.info()[["sysname"]]`.
#'
#' @details
#' This function is used with testthat's `variant` parameter in
#' [testthat::expect_snapshot_file()] to create OS-specific snapshots
#' when file formats produce different output across platforms.
#'
#' Common use cases:
#' - RDS files: Binary serialization format that can differ across OS
#' - Platform-specific numeric computations (e.g., MCMC algorithms)
#' - Files with platform-specific line endings or encodings
#'
#' @returns A character string: "linux", "darwin" (macOS), or "windows"
#' @export
#' @examples
#' system_os()
system_os <- function() {
  tolower(Sys.info()[["sysname"]])
}

#' Get darwin snapshot variant for macOS
#'
#' @description
#' Returns "darwin" for macOS, NULL for other platforms (Linux/Windows).
#'
#' @details
#' This is a specialized variant function for cases where only macOS
#' produces different results while Linux and Windows are identical.
#' Use this instead of [system_os()] when:
#' - macOS produces unique output (e.g., platform-specific math libraries)
#' - Linux and Windows produce identical results
#' - You want to maintain a single snapshot for Linux/Windows
#'
#' Common use cases:
#' - JAGS MCMC output (macOS uses different floating-point arithmetic)
#' - Platform-specific numerical computations
#'
#' @returns "darwin" on macOS, NULL on other platforms
#' @export
#' @examples
#' darwin_variant()
darwin_variant <- function() {
  if (system_os() == "darwin") "darwin" else NULL
}

#' Get platform variant including OS and R version
#'
#' @description
#' Returns a variant string combining OS and R major.minor version.
#' This is useful for RDS snapshots that vary by both platform and R version.
#'
#' @details
#' RDS files can produce different binary output across R versions even when
#' using the same serialization version. This function creates variant strings
#' like "linux-4.4", "darwin-4.5", "windows-4.4" to handle these differences.
#'
#' Use this function with [testthat::expect_snapshot_file()] when testing
#' RDS files or other formats that vary by both OS and R version.
#'
#' @returns A character string like "linux-4.4", "darwin-4.5", or
#'   "windows-4.4"
#' @export
#' @examples
#' platform_variant()
platform_variant <- function() {
  # Get R version as major.minor (e.g., "4.4" for R 4.4.3)
  # Patch versions shouldn't affect RDS serialization
  r_ver <- getRversion()
  r_version <- paste(r_ver$major, r_ver$minor, sep = ".")
  paste0(system_os(), "-", r_version)
}

#' Snapshot testing for [data.frame]s
#' @description
#' copied from <https://github.com/bcgov/ssdtools>
#' with permission (<https://github.com/bcgov/ssdtools/issues/379>)
#'
#' @param x a [data.frame] to snapshot
#' @param name [character] snapshot name
#' @param digits [integer] passed to [signif()] for numeric variables
#' @inheritDotParams testthat::expect_snapshot_file -path -name -compare
#' @returns [NULL] (from [testthat::expect_snapshot_file()])
#' @export
#' @examples
#' \donttest{
#' # expect_snapshot_data() must be called inside a test_that() block with
#' # testthat 3rd edition active. Outside a test suite, the snapshot is
#' # skipped because there is no reference file to compare against.
#' withr::with_tempdir({
#'   testthat::test_that("iris snapshot", {
#'     testthat::local_edition(3)
#'     expect_snapshot_data(iris, name = "iris")
#'   })
#' })
#' }
expect_snapshot_data <- function(x, name, digits = 6, ...) {
  fun <- function(x) signif(x, digits = digits)
  lapply_fun <- function(x) I(lapply(x, fun))
  x <- dplyr::mutate(x, dplyr::across(tidyselect::where(is.numeric), fun))
  x <- dplyr::mutate(x, dplyr::across(tidyselect::where(is.list), lapply_fun))
  expect_snapshot_object(x, name = name, writer = save_csv, ...)
}

#' Save a data.frame to a CSV file
#' @param x A data.frame to save
#' @returns [character] Path to the temporary CSV file
#' @keywords internal
#' @export
save_csv <- function(x) {
  path <- tempfile(fileext = ".csv")
  readr::write_csv(x, path)
  path
}

---
title: "Getting Started with snapr"
vignette: >
  %\VignetteIndexEntry{Getting Started with snapr}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

## Introduction

Snapshot testing captures the output of a function or computation and saves it
to a file. On subsequent test runs, the output is compared to the saved
snapshot. If the output changes, the test fails — alerting you to unexpected
behaviour.

`snapr` builds on top of [testthat](https://testthat.r-lib.org/)'s
`expect_snapshot_file()` to provide convenient helpers for common R objects:

- `expect_snapshot_data()` — for data.frames (stores as CSV for
  human-readable diffs)
- `expect_snapshot_object()` — for any R object (stores as RDS, JSON, or
  plain text depending on the `writer` you choose)

## Installation

Install the development version from GitHub:

```r
# install.packages("pak")
pak::pak("d-morrison/snapr")
```

## Basic Workflow

Snapshot tests live inside `test_that()` blocks, just like any other
[testthat](https://testthat.r-lib.org/) expectation.

### Snapshot a data.frame

Use `expect_snapshot_data()` when you want a human-readable CSV snapshot:

```r
library(snapr)

test_that("iris subset is stable", {
  expect_snapshot_data(iris[1:5, ], name = "iris_head")
})
```

The first time you run this test, `snapr` creates
`tests/testthat/_snaps/iris_head.csv`. On subsequent runs the data.frame is
compared against that file and the test fails if anything has changed.

### Snapshot any R object

Use `expect_snapshot_object()` for objects that cannot be easily represented
as a CSV. The default writer (`save_rds()`) stores the object as an `.rds`
file and uses [diffobj](https://github.com/brodieG/diffobj) to produce
human-readable diffs during review:

```r
test_that("model structure is stable", {
  model <- lm(mpg ~ wt, data = mtcars)
  expect_snapshot_object(model, name = "mpg_model")
})
```

## Choosing a Writer

The `writer` argument of `expect_snapshot_object()` controls the file format.
`snapr` ships four built-in writers:

| Writer | Format | Best for |
|--------|--------|----------|
| `save_rds()` | `.rds` (binary) | Any R object; rich diffs via diffobj |
| `save_json()` | `.json` (text) | Lists and data that map cleanly to JSON |
| `save_csv()` | `.csv` (text) | data.frames (also used by `expect_snapshot_data()`) |
| `save_deparse()` | `.txt` (text) | Simple objects; always human-readable |

### JSON example

```r
test_that("config list is stable", {
  config <- list(name = "my_app", version = "1.0.0", debug = TRUE)
  expect_snapshot_object(config, name = "config", writer = save_json)
})
```

### Deparse example

```r
test_that("simple list is stable", {
  expect_snapshot_object(list(x = 1:3), name = "simple", writer = save_deparse)
})
```

## Handling Platform Differences

Some R objects (particularly RDS files and results from numerical algorithms)
can differ across operating systems or R versions. `snapr` provides helper
functions to create *variant* strings that are appended to snapshot names,
keeping snapshots separate per platform:

| Function | Variant string | Use when… |
|----------|---------------|-----------|
| `system_os()` | `"linux"`, `"darwin"`, `"windows"` | Output differs by OS only |
| `darwin_variant()` | `"darwin"` / `NULL` | Only macOS differs; Linux & Windows share one snapshot |
| `platform_variant()` | `"linux-4.4"` etc. | Output differs by both OS *and* R version |

```r
test_that("RDS snapshot is platform-safe", {
  result <- list(x = rnorm(3))
  expect_snapshot_object(
    result,
    name = "result",
    variant = platform_variant()
  )
})
```

Text-based formats (JSON, CSV, deparse) typically produce identical output
across platforms and do not require a variant.

## Reviewing Snapshots

When a snapshot test fails, you can interactively review the differences with:

```r
testthat::snapshot_review()
```

This opens a Shiny app that displays the old and new snapshots side-by-side,
letting you accept or reject each change.

For richer, visual diffs when reviewing RDS snapshots, install the
[d-morrison/diffviewer](https://github.com/d-morrison/diffviewer) fork of
[diffviewer](https://github.com/r-lib/diffviewer). This fork wraps
[diffobj](https://github.com/brodieG/diffobj) to compare deserialized R
objects with arbitrary structures rather than raw bytes:

```r
pak::pak("d-morrison/diffviewer")
```

To accept all new snapshots at once (e.g., after an intentional change):

```r
testthat::snapshot_accept()
```

## Summary

| Task | Function |
|------|----------|
| Snapshot a data.frame | `expect_snapshot_data(x, name)` |
| Snapshot any object (RDS) | `expect_snapshot_object(x, name)` |
| Snapshot as JSON | `expect_snapshot_object(x, name, writer = save_json)` |
| Snapshot as CSV | `expect_snapshot_object(x, name, writer = save_csv)` |
| Snapshot as plain text | `expect_snapshot_object(x, name, writer = save_deparse)` |
| OS-specific variant | `variant = system_os()` |
| macOS-only variant | `variant = darwin_variant()` |
| OS + R-version variant | `variant = platform_variant()` |
| Review failing snapshots | `testthat::snapshot_review()` |
| Accept new snapshots | `testthat::snapshot_accept()` |

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare_file_object.R
\name{compare_file_object}
\alias{compare_file_object}
\title{Compare RDS files using waldo for better snapshot review}
\usage{
compare_file_object(old, new, print = FALSE, ...)
}
\arguments{
\item{old}{Path to the old (reference) RDS file}

\item{new}{Path to the new RDS file to compare}

\item{print}{\link{logical} whether to print \link[waldo:compare]{waldo::compare} output to R console;
can become very long for complex objects like \link{lm}s}

\item{...}{
  Arguments passed on to \code{\link[waldo:compare]{waldo::compare}}
  \describe{
    \item{\code{x,y}}{Objects to compare. \code{x} is treated as the reference object
so messages describe how \code{y} is different to \code{x}.}
    \item{\code{x_arg,y_arg}}{Name of \code{x} and \code{y} arguments, used when generated paths
to internal components. These default to "old" and "new" since it's
most natural to supply the previous value then the new value.}
    \item{\code{tolerance}}{If non-\code{NULL}, used as threshold for ignoring small
floating point difference when comparing numeric vectors. Using any
non-\code{NULL} value will cause integer and double vectors to be compared
based on their values, not their types, and will ignore the difference
between \code{NaN} and \code{NA_real_}.

It uses the same algorithm as \code{\link[=all.equal]{all.equal()}}, i.e., first we generate
\code{x_diff} and \code{y_diff} by subsetting \code{x} and \code{y} to look only locations
with differences. Then we check that
\code{mean(abs(x_diff - y_diff)) / mean(abs(y_diff))} (or just
\code{mean(abs(x_diff - y_diff))} if \code{y_diff} is small) is less than
\code{tolerance}.}
    \item{\code{max_diffs}}{Control the maximum number of differences shown. The
default shows 10 differences when run interactively and all differences
when run in CI. Set \code{max_diffs = Inf} to see all differences.}
    \item{\code{ignore_srcref}}{Ignore differences in function \code{srcref}s? \code{TRUE} by
default since the \code{srcref} does not change the behaviour of a function,
only its printed representation.}
    \item{\code{ignore_attr}}{Ignore differences in specified attributes?
Supply a character vector to ignore differences in named attributes.
By default the \code{"waldo_opts"} attribute is listed in \code{ignore_attr} so
that changes to it are not reported; if you customize \code{ignore_attr}, you
will probably want to do this yourself.

For backward compatibility with \code{all.equal()}, you can also use \code{TRUE},
to all ignore differences in all attributes. This is not generally
recommended as it is a blunt tool that will ignore many important
functional differences.}
    \item{\code{ignore_encoding}}{Ignore string encoding? \code{TRUE} by default, because
this is R's default behaviour. Use \code{FALSE} when specifically concerned
with the encoding, not just the value of the string.}
    \item{\code{ignore_function_env,ignore_formula_env}}{Ignore the environments of
functions and formulas, respectively? These are provided primarily for
backward compatibility with \code{all.equal()} which always ignores these
environments.}
    \item{\code{list_as_map}}{Compare lists as if they are mappings between names and
values. Concretely, this drops \code{NULL}s in both objects and sorts named
components.}
    \item{\code{quote_strings}}{Should strings be surrounded by quotes? If \code{FALSE},
only side-by-side and line-by-line comparisons will be used, and there's
no way to distinguish between \code{NA} and \code{"NA"}.}
  }}
}
\value{
\link{logical} TRUE if objects are identical, FALSE otherwise
}
\description{
This comparison function loads RDS files and compares the deserialized
R objects rather than raw bytes. This enables better visualization in
snapshot_review() by comparing the actual object structure rather than
binary serialization.

For meaningful diffs in testthat's snapshot_review(), this function
compares the R objects after deserialization using \code{\link[waldo:compare]{waldo::compare()}}.
}
\examples{
old_obj <- list(a = 1, b = 2)
new_obj <- list(a = 1, b = 3)
old_path <- tempfile(fileext = ".rds")
new_path <- tempfile(fileext = ".rds")
saveRDS(old_obj, old_path)
saveRDS(new_obj, new_path)
compare_file_object(old_path, new_path)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect_snapshot_data.R
\name{expect_snapshot_data}
\alias{expect_snapshot_data}
\title{Snapshot testing for \link{data.frame}s}
\usage{
expect_snapshot_data(x, name, digits = 6, ...)
}
\arguments{
\item{x}{a \link{data.frame} to snapshot}

\item{name}{\link{character} snapshot name}

\item{digits}{\link{integer} passed to \code{\link[=signif]{signif()}} for numeric variables}

\item{...}{
  Arguments passed on to \code{\link[testthat:expect_snapshot_file]{testthat::expect_snapshot_file}}
  \describe{
    \item{\code{binary}}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Please use the
\code{compare} argument instead.}
    \item{\code{cran}}{Should these expectations be verified on CRAN? By default,
they are not, because snapshot tests tend to be fragile because they
often rely on minor details of dependencies.}
    \item{\code{transform}}{Optionally, a function to scrub sensitive or stochastic
text from the output. Should take a character vector of lines as input
and return a modified character vector as output.}
    \item{\code{variant}}{If not-\code{NULL}, results will be saved in
\verb{_snaps/{variant}/{test}/{name}}. This allows you to create
different snapshots for different scenarios, like different operating
systems or different R versions.

Note that there's no way to declare all possible variants up front which
means that as soon as you start using variants, you are responsible for
deleting snapshot variants that are no longer used. (testthat will still
delete all variants if you delete the test.)}
    \item{\code{old,new}}{Paths to old and new snapshot files.}
  }}
}
\value{
\link{NULL} (from \code{\link[testthat:expect_snapshot_file]{testthat::expect_snapshot_file()}})
}
\description{
copied from \url{https://github.com/bcgov/ssdtools}
with permission (\url{https://github.com/bcgov/ssdtools/issues/379})
}
\examples{
\donttest{
# expect_snapshot_data() must be called inside a test_that() block with
# testthat 3rd edition active. Outside a test suite, the snapshot is
# skipped because there is no reference file to compare against.
withr::with_tempdir({
  testthat::test_that("iris snapshot", {
    testthat::local_edition(3)
    expect_snapshot_data(iris, name = "iris")
  })
})
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect_snapshot_object.R
\name{expect_snapshot_object}
\alias{expect_snapshot_object}
\title{Snapshot testing for R objects}
\usage{
expect_snapshot_object(
  x,
  name,
  writer = save_rds,
  print = FALSE,
  tolerance = NULL,
  ...
)
}
\arguments{
\item{x}{An R object to snapshot. Can be any R object including lists,
models, data.frames, vectors, etc.}

\item{name}{\link{character} snapshot name (file extension added
automatically)}

\item{writer}{\link{function} Function to write the object to a file.
Default is \code{\link[=save_rds]{save_rds()}}.
Other options include \code{\link[=save_json]{save_json()}}, \code{\link[=save_deparse]{save_deparse()}}, \code{\link[=save_csv]{save_csv()}}.
Custom writer functions should accept \code{x} and return a file path.}

\item{print}{\link{logical} whether to print \link[waldo:compare]{waldo::compare} output to R console;
can become very long for complex objects like \link{lm}s}

\item{tolerance}{If non-\code{NULL}, used as threshold for ignoring small
floating point differences when comparing numeric vectors. Only applies
when \code{writer} produces an RDS file (the default); silently ignored for
text-based formats (JSON, CSV, deparse, etc.). See \code{\link[waldo:compare]{waldo::compare()}} for
full details.}

\item{...}{
  Arguments passed on to \code{\link[testthat:expect_snapshot_file]{testthat::expect_snapshot_file}}
  \describe{
    \item{\code{binary}}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Please use the
\code{compare} argument instead.}
    \item{\code{cran}}{Should these expectations be verified on CRAN? By default,
they are not, because snapshot tests tend to be fragile because they
often rely on minor details of dependencies.}
    \item{\code{transform}}{Optionally, a function to scrub sensitive or stochastic
text from the output. Should take a character vector of lines as input
and return a modified character vector as output.}
    \item{\code{variant}}{If not-\code{NULL}, results will be saved in
\verb{_snaps/{variant}/{test}/{name}}. This allows you to create
different snapshots for different scenarios, like different operating
systems or different R versions.

Note that there's no way to declare all possible variants up front which
means that as soon as you start using variants, you are responsible for
deleting snapshot variants that are no longer used. (testthat will still
delete all variants if you delete the test.)}
    \item{\code{old,new}}{Paths to old and new snapshot files.}
  }}
}
\value{
\link{NULL} (from \code{\link[testthat:expect_snapshot_file]{testthat::expect_snapshot_file()}})
}
\description{
A flexible wrapper around \code{\link[testthat:expect_snapshot_file]{testthat::expect_snapshot_file()}} that allows
snapshotting any R object, not just data.frames. This function provides
a convenient interface for snapshot testing with sensible defaults for
serialization.

When using RDS format (the default), snapshots are compared using
\code{\link[waldo:compare]{waldo::compare()}} which provides rich, visual diffs in
\code{\link[testthat:snapshot_review]{testthat::snapshot_review()}}. This makes it much easier to review
changes to complex R objects.
}
\details{
When using RDS format (the default), snapshots can vary across R versions
and platforms even with fixed serialization versions. Consider using
\code{variant = platform_variant()} for RDS snapshots to handle these
differences, or use text-based formats like JSON or deparse for more
stable snapshots across platforms and versions.

The RDS comparison uses \code{\link[waldo:compare]{waldo::compare()}} internally, which provides
rich visual diffs in \code{\link[testthat:snapshot_review]{testthat::snapshot_review()}}. This is particularly
useful for complex objects like models, nested lists, or data structures
where byte-level comparison would be difficult to interpret.
}
\examples{
\donttest{
# expect_snapshot_object() must be called inside a test_that() block with
# testthat 3rd edition active. Outside a test suite, the snapshot is
# skipped because there is no reference file to compare against.
withr::with_tempdir({
  testthat::test_that("snapshot examples", {
    testthat::local_edition(3)

    # Snapshot a list (RDS format with platform/version variant)
    expect_snapshot_object(
      list(a = 1, b = 2), name = "config", variant = platform_variant()
    )

    # Snapshot a model
    model <- lm(mpg ~ wt, data = mtcars)
    expect_snapshot_object(
      model, name = "model", variant = platform_variant()
    )

    # Snapshot with JSON format (for human-readable diffs)
    # Text formats don't need variants
    expect_snapshot_object(iris[1:5, ], name = "iris", writer = save_json)

    # Snapshot with deparse format
    expect_snapshot_object(
      list(x = 1:5), name = "simple_list", writer = save_deparse
    )
  })
})
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system_os.R
\name{system_os}
\alias{system_os}
\title{Get the current operating system}
\usage{
system_os()
}
\value{
A character string: "linux", "darwin" (macOS), or "windows"
}
\description{
Returns the name of the current operating system in lowercase.
This is a simple wrapper around \code{Sys.info()[["sysname"]]}.
}
\details{
This function is used with testthat's \code{variant} parameter in
\code{\link[testthat:expect_snapshot_file]{testthat::expect_snapshot_file()}} to create OS-specific snapshots
when file formats produce different output across platforms.

Common use cases:
\itemize{
\item RDS files: Binary serialization format that can differ across OS
\item Platform-specific numeric computations (e.g., MCMC algorithms)
\item Files with platform-specific line endings or encodings
}
}
\examples{
system_os()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect_snapshot_object.R
\name{save_deparse}
\alias{save_deparse}
\title{Save an R object to a text file using deparse}
\usage{
save_deparse(x)
}
\arguments{
\item{x}{An R object to save}
}
\value{
\link{character} Path to the temporary text file
}
\description{
Save an R object to a text file using deparse
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system_os.R
\name{darwin_variant}
\alias{darwin_variant}
\title{Get darwin snapshot variant for macOS}
\usage{
darwin_variant()
}
\value{
"darwin" on macOS, NULL on other platforms
}
\description{
Returns "darwin" for macOS, NULL for other platforms (Linux/Windows).
}
\details{
This is a specialized variant function for cases where only macOS
produces different results while Linux and Windows are identical.
Use this instead of \code{\link[=system_os]{system_os()}} when:
\itemize{
\item macOS produces unique output (e.g., platform-specific math libraries)
\item Linux and Windows produce identical results
\item You want to maintain a single snapshot for Linux/Windows
}

Common use cases:
\itemize{
\item JAGS MCMC output (macOS uses different floating-point arithmetic)
\item Platform-specific numerical computations
}
}
\examples{
darwin_variant()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect_snapshot_data.R
\name{save_csv}
\alias{save_csv}
\title{Save a data.frame to a CSV file}
\usage{
save_csv(x)
}
\arguments{
\item{x}{A data.frame to save}
}
\value{
\link{character} Path to the temporary CSV file
}
\description{
Save a data.frame to a CSV file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapr-package.R
\docType{package}
\name{snapr-package}
\alias{snapr}
\alias{snapr-package}
\title{snapr: Convenient Snapshot Testing Functions for Packages}
\description{
\if{html}{\figure{logo.svg}{options: style='float: right' alt='logo' width='120'}}

Provides convenient snapshot testing functions for packages, including expect_snapshot_data() for data.frames and expect_snapshot_object() for any R object.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/d-morrison/snapr}
  \item \url{https://d-morrison.github.io/snapr/}
  \item Report bugs at \url{https://github.com/d-morrison/snapr/issues}
}

}
\author{
\strong{Maintainer}: Douglas Ezra Morrison \email{demorrison@ucdavis.edu}

Authors:
\itemize{
  \item Douglas Ezra Morrison \email{demorrison@ucdavis.edu}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system_os.R
\name{platform_variant}
\alias{platform_variant}
\title{Get platform variant including OS and R version}
\usage{
platform_variant()
}
\value{
A character string like "linux-4.4", "darwin-4.5", or
"windows-4.4"
}
\description{
Returns a variant string combining OS and R major.minor version.
This is useful for RDS snapshots that vary by both platform and R version.
}
\details{
RDS files can produce different binary output across R versions even when
using the same serialization version. This function creates variant strings
like "linux-4.4", "darwin-4.5", "windows-4.4" to handle these differences.

Use this function with \code{\link[testthat:expect_snapshot_file]{testthat::expect_snapshot_file()}} when testing
RDS files or other formats that vary by both OS and R version.
}
\examples{
platform_variant()
}
